module Exception (
	Exception, catchException, throwException, mayCatchException,
	errorToException
    ) where

import Control.Monad.Trans
import Control.Monad.Error

-- The monad m can throw exceptions of type a.
class (Monad m) => Exception a m where
    mayCatchException :: m b -> (a -> Maybe (m b)) -> m b
    throwException :: a -> m t

-- The exception monad built by adding status type 'e' to monad 'k'
-- can raise exceptions of type 'e'.
instance (Monad m, Error e) => Exception e (ErrorT e m) where
    --	ErrorT e m b -> (e -> Maybe (ErrorT e m b)) -> ErrorT e m b
    mayCatchException (ErrorT act) handler = ErrorT $ do
	res <- act
	case res of
	    Left err -> case handler err of
			    Just (ErrorT act) -> act
			    Nothing -> return res
	    Right val -> return res
    -- raise :: (Monad m) => s -> ErrorT e m t
    throwException err = ErrorT $ return $ Left $ err

catchException :: (Exception a m) => m b -> (a -> m b) -> m b
catchException m handler = mayCatchException m (Just . handler)

errorToException :: (Monad m) => Either e t -> ErrorT e m t
errorToException v = ErrorT (return v)
