module FTPError (
	FTPException,
        ftp, cwd, retrbinary, write, quit, FTPName, makeFTPName,
        ErrorPerm(ErrorPerm),
        ErrorTemp(ErrorTemp),
        FTPError(FTPError))
    where
import Exception
import Control.Monad.Trans
import Control.Monad.Error
import AnyError
import Subexception

-- FTPError
data FTPError = FTPError String | ErrorFTPPerm ErrorPerm | ErrorFTPTemp ErrorTemp
	deriving (Show)
-- is a subexception of AnyError
instance Subexception FTPError AnyError AnyException where
	inject err = case err of
		FTPError msg -> AnyError msg
		ErrorFTPPerm (ErrorPerm msg) -> AnyError msg
		ErrorFTPTemp (ErrorTemp msg) -> AnyError msg
instance Error FTPError where
	strMsg = FTPError

-- ErrorPerm
data ErrorPerm = ErrorPerm String
	deriving (Show)
-- is an Exception type.
instance Exception ErrorPerm FTPException where
	mayCatchException act handler = mayCatchException act (\eftp -> case eftp of
						ErrorFTPPerm ep -> handler ep
						otherwise -> Nothing)
	throwException = throwException . ErrorFTPPerm

-- ErrorTemp
data ErrorTemp = ErrorTemp String
	deriving (Show)
-- is an exception type.
instance Exception ErrorTemp FTPException where
	mayCatchException act handler = mayCatchException act (\eftp -> case eftp of
						ErrorFTPTemp ep -> handler ep
						otherwise -> Nothing)
	throwException = throwException . ErrorFTPTemp

type FTPException = ErrorT FTPError AnyException

ftp :: String -> FTPException ()
ftp str = if take 3 str == "ftp"
        then lift (return ())
        else throwException $ ErrorPerm ("ftp " ++ str)

cwd :: String -> FTPException ()
cwd str = if take 1 str == "/"
        then lift (return ())
        else throwException $ FTPError ("cwd " ++ str)

data Writeable = Writeable

retrbinary :: String -> (Writeable -> FTPException ()) -> FTPException ()
retrbinary str f = f Writeable

write :: Writeable -> FTPException ()
write x = return ()

quit :: FTPException ()
quit = return ()

-- Example of integrating the Error (Either) monad with exceptions.
data FTPName = FTPName String deriving (Show)

makeFTPName :: String -> Either FTPError FTPName
makeFTPName x = if null x then throwError (FTPError "empty name")
		 else return (FTPName x)
