module Subexception (
    Subexception, inject,
    runException
) where

import System.Exit
import Exception
import Control.Monad.Error

-- Subexception class
class (Monad m, Exception e m) => Subexception s e m | s -> e, s -> m where
    inject :: s -> e

-- A Subexception can handle exceptions in two categories:
-- those it is created to handle directly, and those which
-- it can pass along to the exception monad it is derived from.
-- The former case is listed in the Exception module.

try :: (Exception e m) => m b -> m (Either e b)
try act = (do
		res1 <- act
		return $ Right res1)
	 `catchException` (return . Left)

instance (Error s, Subexception s e eem) =>
	   Exception e (ErrorT s eem)
  where
    -- mayCatchException :: ErrorT s eem t ->
    --		       (e -> Maybe (ption s eem t)) ->
    --			ErrorT s eem t
    mayCatchException (ErrorT act :: ErrorT s eem t)
	   handler -- (handler  :: e -> Maybe (ErrorT s eem t))
	= ErrorT $ do
		res0 <- try act  -- :: eem (Either e (Either s t))
		(act0 :: Either s t) <- case res0 of
		    Left (err :: e) -> case handler err of
					Just (ErrorT act) -> (act :: eem (Either s t))
					Nothing -> throwException err
		    Right (res1:: Either s t) -> case res1 of
			Left (err :: s) -> case handler (inject err) of
					Just (ErrorT act) -> act
					Nothing -> return res1
			Right _ -> return res1
		(return act0 :: eem (Either s t))

    -- throwException :: e -> ErrorT s eem t
    throwException err = ErrorT $ throwException err

runException :: (Subexception s e m) => ErrorT s m a -> m a
runException (ErrorT thrower) = do
	m <- thrower
	case m of
	    Left err -> throwException $ inject err
	    Right r -> return r
